from functools import wraps

import pandas as pd
from frictionless.console.common import source

from pydeflate.core.api import BaseExchange
from pydeflate.core.source import DAC, WorldBank, IMF, WorldBankPPP


def _generate_docstring(source_name: str) -> str:
    """Generate docstring for each decorated exchange function."""
    return (
        f"Exchange a DataFrame using the {source_name} rates source.\n\n"
        f"This function applies exchange rates toa  DataFrame using the {source_name} rates.\n\n"
        "Args:\n"
        "    data (pd.DataFrame): The input DataFrame containing data to deflate.\n"
        "    source_currency (str, optional): The source currency code. Defaults to 'USA'.\n"
        "    target_currency (str, optional): The target currency code. Defaults to 'USA'.\n"
        "    id_column (str, optional): Column with entity identifiers. Defaults to 'iso_code'.\n"
        "    year_column (str, optional): Column with year information. Defaults to 'year'.\n"
        "    use_source_codes (bool, optional): Use source-specific entity codes. Defaults to False.\n"
        "    value_column (str, optional): Column with values to deflate. Defaults to 'value'.\n"
        "    target_value_column (str, optional): Column to store deflated values. Defaults to 'value'.\n"
        "    reversed_ (bool, optional): The reverse of an exchange conversion. Defaults to False.\n"
        "    year_format (str | None, optional): Format of the year in `year_column`. Defaults to None.\n"
        "    update_rates (bool, optional): Update the exchange rate data. Defaults to False.\n\n"
        "Returns:\n"
        "    pd.DataFrame: DataFrame with converted values in the `target_value_column`.\n"
    )


def _exchange(exchange_source_cls, **fixed_params):
    """Decorator to create exchange wrappers with specific source"""

    def decorator(func):
        @wraps(func)
        def wrapper(data: pd.DataFrame, **kwargs):
            # Validate input parameters
            for param in fixed_params:
                if param in kwargs:
                    raise ValueError(
                        f"The parameter '{param}' cannot be passed to this function."
                    )
            # set fixed parameters
            kwargs.update(fixed_params)

            # Unpack the parameters
            source_currency = kwargs.get("source_currency", "USA")
            target_currency = kwargs.get("target_currency", "USA")
            id_column = kwargs.get("id_column", "iso_code")
            year_column = kwargs.get("year_column", "year")
            use_source_codes = kwargs.get("use_source_codes", False)
            value_column = kwargs.get("value_column", "value")
            target_value_column = kwargs.get("target_value_column", "value")
            reversed_ = kwargs.get("reversed_", False)
            year_format = kwargs.get("year_format", None)
            update_rates = kwargs.get("update_rates", False)

            if not isinstance(data, pd.DataFrame):
                raise ValueError("The 'data' parameter must be a pandas DataFrame.")

            if id_column not in data.columns:
                raise ValueError(
                    f"The id_column '{id_column}' is not in the DataFrame."
                )
            if year_column not in data.columns:
                raise ValueError(
                    f"The year_column '{year_column}' is not in the DataFrame."
                )
            if value_column not in data.columns:
                raise ValueError(
                    f"The value_column '{value_column}' is not in the DataFrame."
                )

            # Copy the data to avoid modifying the original
            to_exchange = data.copy()

            # Initialize the deflator source
            if exchange_source_cls.__name__ == "WorldBankPPP":
                source = exchange_source_cls(
                    update=update_rates,
                    from_lcu=False if source_currency == "USA" else True,
                )
                source_currency = "LCU" if source_currency == "USA" else source_currency
            else:
                source = exchange_source_cls(update=update_rates)

            # Create a deflator object
            exchange = BaseExchange(
                exchange_source=source,
                source_currency=source_currency,
                target_currency=target_currency,
                use_source_codes=use_source_codes,
            )

            # Deflate the data
            return exchange.exchange(
                data=to_exchange,
                entity_column=id_column,
                year_column=year_column,
                value_column=value_column,
                target_value_column=target_value_column,
                year_format=year_format,
                reversed_=reversed_,
            )

        # Add the deflator source and price kind to the function
        wrapper.__doc__ = _generate_docstring(exchange_source_cls.__name__)
        return wrapper

    return decorator


@_exchange(DAC)
def oecd_dac_exchange(
    data: pd.DataFrame,
    *,
    source_currency: str = "USA",
    target_currency: str = "USA",
    id_column: str = "iso_code",
    year_column: str = "year",
    use_source_codes: bool = False,
    value_column: str = "value",
    target_value_column: str = "value",
    reversed_: bool = False,
    year_format: str | None = None,
    update_rates: bool = False,
) -> pd.DataFrame: ...


@_exchange(WorldBank)
def wb_exchange(
    data: pd.DataFrame,
    *,
    source_currency: str = "USA",
    target_currency: str = "USA",
    id_column: str = "iso_code",
    year_column: str = "year",
    use_source_codes: bool = False,
    value_column: str = "value",
    target_value_column: str = "value",
    reversed_: bool = False,
    year_format: str | None = None,
    update_rates: bool = False,
) -> pd.DataFrame: ...


@_exchange(WorldBankPPP, target_currency="PPP")
def wb_exchange_ppp(
    data: pd.DataFrame,
    *,
    source_currency: str = "USA",
    id_column: str = "iso_code",
    year_column: str = "year",
    use_source_codes: bool = False,
    value_column: str = "value",
    target_value_column: str = "value",
    reversed_: bool = False,
    year_format: str | None = None,
    update_rates: bool = False,
) -> pd.DataFrame: ...


@_exchange(IMF)
def imf_exchange(
    data: pd.DataFrame,
    *,
    source_currency: str = "USA",
    target_currency: str = "USA",
    id_column: str = "iso_code",
    year_column: str = "year",
    use_source_codes: bool = False,
    value_column: str = "value",
    target_value_column: str = "value",
    reversed_: bool = False,
    year_format: str | None = None,
    update_rates: bool = False,
) -> pd.DataFrame: ...



__author__ = """Jorge Rivera"""
__version__ = "2.1.1"

from pydeflate.deflate.deflators import (
    oecd_dac_deflate,
    wb_cpi_deflate,
    wb_gdp_deflate,
    wb_gdp_linked_deflate,
    imf_cpi_deflate,
    imf_gdp_deflate,
    imf_cpi_e_deflate,
)

from pydeflate.deflate.legacy_deflate import deflate
from pydeflate.exchange.exchangers import (
    oecd_dac_exchange,
    wb_exchange,
    wb_exchange_ppp,
    imf_exchange,
)
from pydeflate.pydeflate_config import setup_logger


def set_pydeflate_path(path):
    from pathlib import Path
    from pydeflate.pydeflate_config import PYDEFLATE_PATHS

    """Set the path to the data folder."""
    global PYDEFLATE_PATHS

    PYDEFLATE_PATHS.data = Path(path).resolve()


__all__ = [
    "set_pydeflate_path",
    "oecd_dac_deflate",
    "oecd_dac_exchange",
    "wb_cpi_deflate",
    "wb_gdp_deflate",
    "wb_gdp_linked_deflate",
    "wb_exchange",
    "imf_cpi_deflate",
    "imf_gdp_deflate",
    "imf_cpi_e_deflate",
    "imf_exchange",
    "deflate",
]



from pathlib import Path

import pandas as pd
from imf_reader import weo

from pydeflate.pydeflate_config import PYDEFLATE_PATHS, logger
from pydeflate.sources.common import (
    today,
    add_pydeflate_iso3,
    enforce_pyarrow_types,
    compute_exchange_deflator,
    read_data,
    prefix_pydeflate_to_columns,
)

# List of WEO indicators of interest
WEO_INDICATORS: list[str] = [
    "NGDP_D",  # Gross domestic product, deflator
    "PCPI",  # Inflation, average consumer prices
    "PCPIE",  # Inflation, end of period consumer prices
    "PPPEX",  # Implied PPP conversion rate
    "NGDPD",  # Gross domestic product, current prices USD
    "NGDP",  # Gross domestic product, current prices
]


def _filter_indicators(df: pd.DataFrame) -> pd.DataFrame:
    """Filter the data to include only selected WEO indicators.

    Args:
        df (pd.DataFrame): The raw dataframe containing WEO data.

    Returns:
        pd.DataFrame: Filtered dataframe with only the relevant indicators.
    """
    return df.loc[df.CONCEPT_CODE.isin(WEO_INDICATORS)]


def _to_units(df: pd.DataFrame) -> pd.DataFrame:
    """Convert OBS_VALUE using the SCALE_CODE multiplier to get proper units.

    Args:
        df (pd.DataFrame): Dataframe with raw observation values.

    Returns:
        pd.DataFrame: Dataframe with scaled observation values.
    """
    df = df.copy()
    df["OBS_VALUE"] = df["OBS_VALUE"] * df["SCALE_CODE"]
    return df


def _clean_columns(df: pd.DataFrame) -> pd.DataFrame:
    """Rename columns to more readable and consistent names.

    Args:
        df (pd.DataFrame): Dataframe with original IMF column names.

    Returns:
        pd.DataFrame: Dataframe with cleaned and renamed columns.
    """
    df = df.rename(
        columns={
            "REF_AREA_CODE": "entity_code",
            "REF_AREA_LABEL": "entity",
            "LASTACTUALDATE": "estimates_start_after",
            "TIME_PERIOD": "year",
            "OBS_VALUE": "value",
        }
    )
    # Standardize column names to snake_case
    df.columns = df.columns.str.lower().str.replace(" ", "_")
    return df


def _keep_useful_columns(df: pd.DataFrame) -> pd.DataFrame:
    """Retain only the columns that are useful for further analysis.

    Args:
        df (pd.DataFrame): Dataframe with cleaned columns.

    Returns:
        pd.DataFrame: Dataframe with only useful columns.
    """
    cols = [
        "year",
        "entity_code",
        "entity",
        "concept_code",
        "value",
    ]
    return df[cols]


def _pivot_concept_code(df: pd.DataFrame) -> pd.DataFrame:
    """Pivot the concept code column to get a wide format for the data.

    Args:
        df (pd.DataFrame): Dataframe with concept code column.

    Returns:
        pd.DataFrame: Dataframe with concept code pivoted to columns.
    """
    return df.pivot(
        index=[c for c in df.columns if c not in ["concept_code", "value"]],
        columns="concept_code",
        values="value",
    ).reset_index()


def _compute_exchange(df: pd.DataFrame) -> pd.DataFrame:
    """Compute the exchange rate and append it to the original DataFrame.

    This function calculates the exchange rate by dividing the 'NGDP'
     Gross domestic product in local currency) by 'NGDPD' (Gross domestic product in USD).
    It then appends this computed exchange rate to the original DataFrame,
    and the new exchange rate rows are labeled with the 'EXCHANGE'
    concept code.

    Args:
        df (pd.DataFrame): Input DataFrame containing columns with
        'concept_code' as 'NGDPD' and 'NGDP'.

    Returns:
        pd.DataFrame: DataFrame with the computed exchange rate included,
        along with the original data.
    """
    # Filter rows with 'NGDPD' (GDP in USD) and 'NGDP' (GDP in local currency)
    exchange = df.loc[lambda d: d.concept_code.isin(["NGDPD", "NGDP"])]

    # Pivot the data so 'NGDPD' and 'NGDP' become separate columns
    exchange = exchange.pipe(_pivot_concept_code)

    # Remove rows that correspond to 'NGDPD' and 'NGDP' from the original DataFrame
    df = df.loc[lambda d: ~d.concept_code.isin(["NGDPD", "NGDP"])]

    # Compute the exchange rate as NGDP (local currency) divided by NGDPD (USD)
    exchange["value"] = round(exchange["NGDP"] / exchange["NGDPD"], 7)

    # Label the exchange rate with a new concept code 'EXCHANGE'
    exchange["concept_code"] = "EXCHANGE"

    # Drop the original 'NGDPD' and 'NGDP' columns as they are no longer needed
    exchange = exchange.drop(columns=["NGDPD", "NGDP"])

    # Concatenate the original DataFrame with the new exchange rate data
    return pd.concat([df, exchange], ignore_index=True)


def download_weo() -> None:
    """Download the WEO data, process it, and save it to a parquet file."""
    logger.info("Downloading the latest WEO data...")

    # Fetch and process the data through a pipeline of transformations
    df = (
        weo.fetch_data()
        .pipe(_filter_indicators)
        .pipe(_to_units)
        .pipe(_clean_columns)
        .pipe(_keep_useful_columns)
        .pipe(_compute_exchange)
        .pipe(add_pydeflate_iso3, column="entity", from_type="regex")
        .pipe(_pivot_concept_code)
        .pipe(compute_exchange_deflator, base_year_measure="NGDP_D")
        .pipe(prefix_pydeflate_to_columns)
        .pipe(enforce_pyarrow_types)
        .reset_index(drop=True)
    )

    # Get today's date to use as a file suffix
    suffix = today()

    # Save the processed dataframe to parquet format
    df.to_parquet(PYDEFLATE_PATHS.data / f"weo_{suffix}.parquet")

    logger.info(f"Saved WEO data to weo_{suffix}.parquet")


def _find_weo_files_in_path(path: Path) -> list:
    """Find all WEO parquet files in the specified directory.

    Args:
        path (Path): The directory path to search for WEO parquet files.

    Returns:
        list: List of WEO parquet files found in the directory.
    """
    return list(path.glob("weo_*.parquet"))


def read_weo(update: bool = False) -> pd.DataFrame:
    """Read the latest WEO data from parquet files or download fresh data."""
    return read_data(
        file_finder_func=_find_weo_files_in_path,
        download_func=download_weo,
        data_name="WEO",
        update=update,
    )


if __name__ == "__main__":
    # Download the WEO data
    dfi = read_weo(update=True)



from pathlib import Path

import pandas as pd
from oda_reader import download_dac1

from pydeflate.pydeflate_config import PYDEFLATE_PATHS
from pydeflate.sources.common import (
    today,
    add_pydeflate_iso3,
    enforce_pyarrow_types,
    compute_exchange_deflator,
    read_data,
    prefix_pydeflate_to_columns,
)


def _find_dac_files_in_path(path: Path) -> list:
    """Find all DAC parquet files in the specified directory.

    Args:
        path (Path): The directory path to search for DAC parquet files.

    Returns:
        list: List of DAC parquet files found in the directory.
    """
    return list(path.glob("dac_*.parquet"))


def _to_units(df: pd.DataFrame) -> pd.DataFrame:
    """Convert DAC values (in million) to units.

    Args:
        df (pd.DataFrame): Dataframe with raw observation values.

    Returns:
        pd.DataFrame: Dataframe with scaled observation values.
    """
    df = df.copy()
    df["value"] = df["value"] * df["unit_multiplier"]
    return df


def _keep_official_definition_only(df: pd.DataFrame) -> pd.DataFrame:
    query = (
        "(aidtype_code == 1010 & flows_code == 1140 & year <2018 ) | "
        "(aidtype_code == 11010 & flows_code == 1160 & year >=2018)"
    )

    return df.query(query)


def _keep_useful_columns(df: pd.DataFrame) -> pd.DataFrame:
    columns = ["year", "donor_code", "donor_name", "EXCHANGE", "DAC_DEFLATOR"]

    return df.filter(columns)


def _pivot_amount_type(df: pd.DataFrame) -> pd.DataFrame:
    df = df.filter(["year", "donor_code", "donor_name", "amounttype_code", "value"])
    return df.pivot(
        index=[c for c in df.columns if c not in ["amounttype_code", "value"]],
        columns="amounttype_code",
        values="value",
    ).reset_index()


def _compute_exchange(df: pd.DataFrame) -> pd.DataFrame:
    # The values for certain providers should be 1
    df.loc[lambda d: d.donor_code >= 20000, "N"] = df.loc[
        lambda d: d.donor_code >= 20000, "A"
    ]
    df["EXCHANGE"] = round(df["N"] / df["A"], 6).fillna(1)
    return df


def _compute_dac_deflator(df: pd.DataFrame) -> pd.DataFrame:
    df["DAC_DEFLATOR"] = round(100 * df["A"] / df["D"], 6)
    return df


def _compute_dac_gdp_deflator(df: pd.DataFrame) -> pd.DataFrame:
    df["NGDP_D"] = round(df["EXCHANGE_D"] / 100 * df["DAC_DEFLATOR"], 5)

    return df


def _rename_columns(df: pd.DataFrame) -> pd.DataFrame:
    return df.rename(
        columns={
            "donor_code": "entity_code",
            "donor_name": "entity",
        }
    )


def download_dac():
    # Use oda_reader to get the data
    df = download_dac1(
        filters={"measure": ["1010", "11010"], "flow_type": ["1140", "1160"]}
    )

    # Clean the data
    df = (
        df.pipe(_to_units)
        .pipe(_keep_official_definition_only)
        .pipe(_pivot_amount_type)
        .pipe(_compute_exchange)
        .pipe(_compute_dac_deflator)
        .pipe(_keep_useful_columns)
        .pipe(add_pydeflate_iso3, column="donor_name", from_type="regex")
        .pipe(_rename_columns)
        .pipe(compute_exchange_deflator, base_year_measure="DAC_DEFLATOR")
        .pipe(_compute_dac_gdp_deflator)
        .pipe(prefix_pydeflate_to_columns)
        .pipe(enforce_pyarrow_types)
        .reset_index(drop=True)
    )

    # Get today's date to use as a file suffix
    suffix = today()

    # Save the data
    df.to_parquet(PYDEFLATE_PATHS.data / f"dac_{suffix}.parquet")


def read_dac(update: bool = False) -> pd.DataFrame:
    """Read the latest WEO data from parquet files or download fresh data."""
    return read_data(
        file_finder_func=_find_dac_files_in_path,
        download_func=download_dac,
        data_name="DAC",
        update=update,
    )


if __name__ == "__main__":
    df = read_dac(update=True)



import warnings

import pandas as pd
from pandas.util._decorators import deprecate_kwarg

from pydeflate.core.api import BaseDeflate
from pydeflate.core.source import DAC, WorldBank, IMF


@deprecate_kwarg(old_arg_name="method", new_arg_name="deflator_method")
@deprecate_kwarg(old_arg_name="source", new_arg_name="deflator_source")
@deprecate_kwarg(old_arg_name="iso_column", new_arg_name="id_column")
@deprecate_kwarg(old_arg_name="source_col", new_arg_name="source_column")
@deprecate_kwarg(old_arg_name="target_col", new_arg_name="target_column")
def deflate(
    df: pd.DataFrame,
    base_year: int,
    deflator_source: str,
    deflator_method: str,
    exchange_source: str,
    exchange_method: str,
    source_currency: str = "USA",
    target_currency: str = "USA",
    id_column: str = "iso_code",
    id_type: str = "ISO3",
    date_column: str = "date",
    source_column: str = "value",
    target_column: str = "value",
    to_current: bool = False,
) -> pd.DataFrame:
    """DEPRECATED: Use pydeflate's updated API for deflation adjustments.

    This legacy function is provided for backwards compatibility only.
    Please refer to the latest documentation for using pydeflate's updated API.

    Args:
        df (pd.DataFrame): DataFrame with flow data to be deflated.
        base_year (int): Target year for base value adjustments.
        deflator_source (str): Source of deflator data, e.g., 'oecd_dac', 'wb', 'imf'.
        deflator_method (str): Method for deflator calculation, e.g., 'gdp', 'cpi'.
        exchange_source (str): Source for exchange rates.
        exchange_method (str): Method for exchange rate calculation.
        source_currency (str): ISO3 code for the source currency.
        target_currency (str): ISO3 code for the target currency.
        id_column (str): Column for entity identifiers, e.g., 'iso_code'.
        id_type (str): Type of ID classification; default is 'ISO3'.
        date_column (str): Column for date information.
        source_column (str): Column with original monetary values.
        target_column (str): Column to store deflated values.
        to_current (bool): If True, convert to current prices.

    Returns:
        pd.DataFrame: DataFrame with deflated values.
    """
    warnings.warn(
        "The `deflate` function is deprecated and will be removed in future versions. "
        "Please check the latest documentation for updated methods.",
        DeprecationWarning,
    )

    if id_type != "ISO3":
        raise ValueError(
            "Only ISO3 ID classification is supported in this version.\n"
            "You can use bblocks to convert to ISO3."
        )

    price_kind = {
        "oecd_dac": "NGDP_D",
        "dac_deflator": "NGDP_D",
        "gdp": "NGDP_D",
        "cpi": "CPI",
    }

    # Mapping of string identifiers to classes and price kinds
    deflator_source_map = {
        "oecd_dac": DAC,
        "dac": DAC,
        "wb": WorldBank,
        "world_bank": WorldBank,
        "imf": IMF,
    }

    deflator_source = deflator_source_map[deflator_source.lower()]()
    exchange_source = deflator_source_map[exchange_source.lower()]()
    deflator_method = price_kind.get(deflator_method.lower(), deflator_method).upper()

    # Copy the data to avoid modifying the original
    to_deflate = df.copy()

    # Create a deflator object
    deflator = BaseDeflate(
        base_year=base_year,
        deflator_source=deflator_source,
        exchange_source=exchange_source,
        source_currency=source_currency,
        target_currency=target_currency,
        price_kind=deflator_method,
        to_current=to_current,
    )

    # Deflate the data
    return deflator.deflate(
        data=to_deflate,
        entity_column=id_column,
        year_column=date_column,
        value_column=source_column,
        target_value_column=target_column,
        year_format=None,
    )



import json

import numpy as np
import pandas as pd

from pydeflate.pydeflate_config import PYDEFLATE_PATHS, logger
from pydeflate.sources.common import enforce_pyarrow_types


def oecd_codes() -> dict:
    with open(PYDEFLATE_PATHS.settings / "oecd_codes.json") as file:
        updates = json.load(file)

    return {int(k): v for k, v in updates.items()}


def emu() -> list:
    with open(PYDEFLATE_PATHS.settings / "emu.json") as file:
        emu = json.load(file)

    return emu


def clean_number(number):
    """Clean a number and return as float"""
    import re

    if not isinstance(number, str):
        number = str(number)

    number = re.sub(r"[^\d.]", "", number)

    if number == "":
        return np.nan

    return float(number)


def create_pydeflate_year(
    data: pd.DataFrame, year_column: str, year_format: str | None = None
) -> pd.DataFrame:
    if year_format is None:
        year_format = "ISO8601"

    data = data.copy()

    data["pydeflate_year"] = pd.to_datetime(
        data[year_column], format=year_format
    ).dt.year

    return data


def _use_implied_dac_rates(
    data: pd.DataFrame,
    pydeflate_data: pd.DataFrame,
    ix: list[str],
    entity_column: str,
    source_codes: bool,
) -> pd.DataFrame:
    """When rates are missing for entities in DAC data, the correct behaviour is to use
    the DAC overall rates"""

    # Assign the DAC code to a temporary column
    data.loc[
        lambda d: ~d[f"temp_{entity_column}"].isin(pydeflate_data[ix[-1]].unique()),
        f"temp_{entity_column}",
    ] = (
        20001 if source_codes else "DAC"
    )

    # Log the fact that implied rates are being used
    flag_missing_pydeflate_data(
        unmatched_data=data.loc[
            lambda d: ~d[f"{entity_column}"].isin(pydeflate_data[ix[-1]].unique())
        ],
        entity_column=entity_column,
        year_column="pydeflate_year",
        using_implied=True,
    )

    return data


def merge_user_and_pydeflate_data(
    data: pd.DataFrame,
    pydeflate_data: pd.DataFrame,
    entity_column: str,
    ix: list[str],
    source_codes: bool = True,
    dac: bool = False,
) -> pd.DataFrame:

    data[f"temp_{entity_column}"] = data[entity_column]

    if dac:
        data = _use_implied_dac_rates(
            data=data,
            pydeflate_data=pydeflate_data,
            ix=ix,
            entity_column=entity_column,
            source_codes=source_codes,
        )

    df_ = data.merge(
        pydeflate_data,
        how="outer",
        left_on=["pydeflate_year", f"temp_{entity_column}"],
        right_on=ix,
        suffixes=("", "_pydeflate"),
        indicator=True,
    ).pipe(enforce_pyarrow_types)

    return df_.drop(columns=[f"temp_{entity_column}"])


def get_unmatched_pydeflate_data(
    merged_data: pd.DataFrame,
):
    return merged_data.loc[merged_data["_merge"] == "left_only"].filter(
        regex="^(?!pydeflate_)(?!.*_pydeflate$)"
    )


def get_matched_pydeflate_data(
    merged_data: pd.DataFrame,
):
    return (
        merged_data.loc[merged_data["_merge"] != "right_only"]
        .drop(columns="_merge")
        .reset_index(drop=True)
    )


def flag_missing_pydeflate_data(
    unmatched_data: pd.DataFrame,
    entity_column: str,
    year_column: str,
    using_implied: bool = False,
):
    """Flag data which is present in the input data but missing in pydeflate's data."""
    if unmatched_data.empty:
        return
    missing = (
        unmatched_data.filter([entity_column, year_column])
        .drop_duplicates()
        .groupby(entity_column)[year_column]
        .apply(lambda x: ", ".join(map(str, sorted(x))))
        .to_dict()
    )

    missing_str = "\n".join(f"{entity}: {years}" for entity, years in missing.items())

    # log all missing data
    message = (
        "Using DAC members' rates (given missing data) for:"
        if using_implied
        else "Missing exchange data for:"
    )
    logger.info(f"{message}\n{missing_str}")



from datetime import datetime
from pathlib import Path
from typing import Any, Literal

import pandas as pd
from hdx.location.country import Country

from pydeflate.pydeflate_config import PYDEFLATE_PATHS, logger

AvailableDeflators = Literal["NGDP_D", "NGDP_DL", "CPI", "PCPI", "PCPIE"]


def check_file_age(file: Path) -> int:
    """Check the age of a WEO file in days.

    Args:
        file (Path): The WEO parquet file to check.

    Returns:
        int: The number of days since the file was created.
    """
    current_date = datetime.today()
    # Extract date from the filename (format: weo_YYYY-MM-DD.parquet)
    file_date = datetime.strptime(file.stem.split("_")[-1], "%Y-%m-%d")

    # Return the difference in days between today and the file's date
    return (current_date - file_date).days


def enforce_pyarrow_types(df: pd.DataFrame) -> pd.DataFrame:
    """Ensures that a DataFrame uses pyarrow dtypes."""
    return df.convert_dtypes(dtype_backend="pyarrow")


def today() -> str:
    from datetime import datetime

    return datetime.today().strftime("%Y-%m-%d")


def _match_regex_to_iso3(
    to_match: list[str], additional_mapping: dict | None
) -> dict[str, str]:
    """Match a list of regex strings to ISO3 country codes.

    Args:
        to_match (list): A list of regex strings to match.

    Returns:
        dict: A dictionary with the regex strings as keys and the ISO3 codes as values.
    """
    if additional_mapping is None:
        additional_mapping = {}

    # Create a Country object
    country = Country()

    # Match the regex strings to ISO3 country codes
    matches = {}

    for match in to_match:
        match_ = country.get_iso3_country_code_fuzzy(match)[0]
        matches[match] = match_
        if match_ is None and match not in additional_mapping:
            logger.debug(f"No ISO3 match found for {match}")

    return matches | additional_mapping


def convert_id(
    series: pd.Series,
    from_type: str = "regex",
    to_type: str = "ISO3",
    not_found: Any = None,
    *,
    additional_mapping: dict = None,
) -> pd.Series:
    """Takes a Pandas' series with country IDs and converts them into the desired type.

    Args:
        series: the Pandas series to convert
        from_type: the classification type according to which the series is encoded.
            For example: ISO3, ISO2, regex, DACCode.
        to_type: the target classification type. Same options as from_type
        not_found: what to do if the value is not found. Can pass a string or None.
            If None, the original value is passed through.
        additional_mapping: Optionally, a dictionary with additional mappings can be used.
            The keys are the values to be converted and the values are the converted values.
            The keys follow the same datatype as the original values. The values must follow
            the same datatype as the target type.
    """

    # if from and to are the same, return without changing anything
    if from_type == to_type:
        return series

    mapping_functions = {"regex": _match_regex_to_iso3}

    # Get the unique values for mapping. This is done in order to significantly improve
    # the performance of country_converter with very long datasets.
    s_unique = series.unique()

    # Create a correspondence dictionary
    mapping = mapping_functions[from_type](
        to_match=s_unique, additional_mapping=additional_mapping
    )

    return series.map(mapping).fillna(series if not_found is None else not_found)


def add_pydeflate_iso3(
    df: pd.DataFrame, column: str, from_type: str = "regex", fillna: Any = pd.NA
) -> pd.DataFrame:
    """Add a column with ISO3 country codes to a dataframe.

    Args:
        df (pd.DataFrame): The dataframe to add the column to.
        column (str): The column containing the country codes.
        from_type (str): The classification type of the country codes.
        fillna (Any): The value to use when the country code is not found.

    Returns:
        pd.DataFrame: The dataframe with the added ISO3 column.
    """
    # Convert the country codes to ISO3
    df["pydeflate_iso3"] = convert_id(
        df[column].fillna(""),
        from_type=from_type,
        to_type="ISO3",
        not_found=fillna,
        additional_mapping={
            "World": "WLD",
            "EU Institutions": "EUI",
            "DAC countries": "DAC",
            "Kosovo": "XXK",
            "G7": "G7C",
            "Sub-Sahara Africa": "SSA",
        },
    )

    return df


def prefix_pydeflate_to_columns(
    df: pd.DataFrame, prefix: str = "pydeflate_"
) -> pd.DataFrame:
    """Add a prefix to all columns in a DataFrame.

    Args:
        df (pd.DataFrame): The DataFrame to add the prefix to.
        prefix (str): The prefix to add to the column names.

    Returns:
        pd.DataFrame: The DataFrame with the prefixed column names.
    """
    df.columns = [
        f"{prefix}{col}" if not col.startswith(prefix) else col for col in df.columns
    ]

    return df


def identify_base_year(df: pd.DataFrame, measure: str, year: str = "year") -> int:
    """Identify the base year for a given measure where the value is equal to 100.

    Args:
        df (pd.DataFrame): DataFrame containing the deflator data with 'year' and the given measure.
        measure (str): The column name for the measure to find the base year for.
        year (str): The column name for the year.

    Returns:
        int: The base year, or None if no base year is found.
    """
    # Find the year where the deflator measure is exactly 100 (or very close)
    base_year = df.loc[df[measure].round(2) == 100, year]

    # Return the year if found, otherwise return None
    return base_year.iloc[0] if not base_year.empty else None


def compute_exchange_deflator(
    df: pd.DataFrame,
    base_year_measure: str | None = None,
    exchange: str = "EXCHANGE",
    year: str = "year",
    grouper: list[str] = None,
) -> pd.DataFrame:
    """Compute the exchange rate deflator for each group of entities.

    This function calculates a deflator for the exchange rate by identifying a base year
    where the base_year_measure is 100, then normalizing exchange values to that base year.

    Args:
        df (pd.DataFrame): Input DataFrame containing columns 'year', and 'EXCHANGE'.
        base_year_measure (str): The column name for the measure to find the base year for.
        exchange (str): The column name for the exchange rate.
        year (str): The column name for the year.
        grouper (list): List of columns to group by before applying the deflator.

    Returns:
        pd.DataFrame: DataFrame with an additional column for the exchange rate deflator.
    """

    def _add_deflator(
        group: pd.DataFrame,
        measure: str | None = "NGDPD_D",
        exchange: str = "EXCHANGE",
        year: str = "year",
    ) -> pd.DataFrame:

        # if needed, clean exchange name
        if exchange.endswith("_to") or exchange.endswith("_from"):
            exchange_name = exchange.rsplit("_", 1)[0]
        else:
            exchange_name = exchange

        # Identify the base year for the deflator
        if measure is not None:
            base_year = identify_base_year(group, measure=measure, year=year)
        else:
            base_year = group.dropna(subset=exchange)[year].max()

        # If no base year is found, return the group unchanged
        if base_year is None or pd.isna(base_year):
            return group

        # Extract the exchange rate value for the base year
        base_value = group.loc[group[year] == base_year, exchange].values

        # If base value is found and valid, calculate the deflator
        if base_value.size > 0 and pd.notna(base_value[0]):
            group[f"{exchange_name}_D"] = round(
                100 * group[exchange] / base_value[0], 6
            )

        return group

    if grouper is None:
        grouper = ["entity", "entity_code"]

    # Apply the deflator computation for each group of 'entity' and 'entity_code'
    return df.groupby(grouper, group_keys=False).apply(
        _add_deflator, measure=base_year_measure, exchange=exchange, year=year
    )


def read_data(
    file_finder_func: callable,
    download_func: callable,
    data_name: str,
    update: bool = False,
) -> pd.DataFrame:
    """Generic function to read data from parquet files or download fresh data.

    Args:
        file_finder_func (function): Function to find existing data files in the path.
        download_func (function): Function to download fresh data if no files are
        found or an update is needed.
        data_name (str): Name of the dataset for logging purposes (e.g., "WEO", "DAC").
        update (bool): If True, forces downloading of new data even if files exist.

    Returns:
        pd.DataFrame: The latest available data.
    """
    # Find existing files using the provided file finder function
    files = file_finder_func(PYDEFLATE_PATHS.data)

    # If no files are found or update is requested, download new data
    if len(files) == 0 or update:
        download_func()
        files = file_finder_func(PYDEFLATE_PATHS.data)

    # If files are found, sort them by age and load the most recent one
    if len(files) > 0:
        files = sorted(files, key=check_file_age)
        latest_file = files[0]

        # Check if the latest file is older than 120 days and log a warning
        if check_file_age(latest_file) > 120:
            logger.warn(
                f"The latest {data_name} data is more than 120 days old.\n"
                f"Consider updating by setting update=True in the function call."
            )

        # Read and return the latest parquet file as a DataFrame
        logger.info(f"Reading {data_name} data from {latest_file}")
        return pd.read_parquet(latest_file)



from dataclasses import dataclass, field
from typing import Literal

import pandas as pd

from pydeflate.core.exchange import Exchange
from pydeflate.core.source import Source
from pydeflate.sources.common import AvailableDeflators


@dataclass
class Deflator:
    """A class to manage and process deflator data.

    This class holds deflators and provides methods to apply exchange rates
    for conversions between source and target currencies.

    """

    source: Source | Exchange
    deflator_type: Literal["price", "exchange"]
    price_kind: AvailableDeflators | None
    base_year: int
    deflator_data: pd.DataFrame = field(default_factory=pd.DataFrame)

    def __post_init__(self):
        """Processes the deflator data based on the deflator type.

        Initializes the `deflator_data` attribute by fetching and processing
        the appropriate deflator data from the source.

        Raises:
            ValueError: If the combination of `deflator_type` and `price_kind` is invalid.
        """

        # If price deflator, fetch the price deflator data
        if self.deflator_type == "price":
            if self.price_kind is None:
                raise ValueError(
                    "`price_kind` must be specified when `deflator_type` is 'price'."
                )
            self.deflator_data = self.source.price_deflator(kind=self.price_kind)

        # If exchange deflator, fetch the exchange deflator data
        elif self.deflator_type == "exchange":
            if self.price_kind is not None:
                raise ValueError(
                    "`price_kind` should be None when `deflator_type` is 'exchange'."
                )
            self.deflator_data = self.source.deflator()
            self.source = self.source.source

        else:
            raise ValueError(f"Invalid deflator type: {self.deflator_type}")

        # Rebase the deflator data to the specified base year
        self.rebase_deflator()

    def _extract_base_year_values(self, value_column: str) -> pd.DataFrame:
        """Extracts the base year values from the deflator data."""

        # Extract base year values
        base_year_values = (
            self.deflator_data[self.deflator_data["pydeflate_year"] == self.base_year]
            .rename(columns={value_column: "base_year_value"})
            .drop(columns=["pydeflate_year"])
        )

        if base_year_values.empty:
            raise ValueError(f"No data found for base year {self.base_year}.")
        return base_year_values

    @staticmethod
    def _ensure_deflator_suffix(value_column: str) -> str:
        """Ensures that the value column ends with the '_D' suffix."""
        return value_column if value_column.endswith("_D") else value_column + "_D"

    def _get_deflator_value_column(self) -> str:
        """Identifies the value column in the deflator data."""
        # Identify the value column
        value_column = self.deflator_data.columns.difference(self.source._idx).tolist()

        if len(value_column) != 1:
            raise ValueError("Invalid deflator data format.")

        return value_column[0]

    def rebase_deflator(self) -> None:
        """Rebases the deflator data to the specified base year.

        Adjusts the deflator values so that the base year values are set to 100.

        Raises:
            ValueError: If `deflator_data` is empty, has an invalid format, or lacks data for the base year.
        """

        # Check if deflator data is empty
        if self.deflator_data.empty:
            raise ValueError("No deflator data found.")

        # Identify the value column
        value_column = self._get_deflator_value_column()

        # Extract base year values
        base_year_values = self._extract_base_year_values(value_column)

        # Merge base year values back into the main DataFrame
        rebased = self.deflator_data.merge(
            base_year_values,
            on=[c for c in self.source._idx if c != "pydeflate_year"],
            how="left",
        )

        # if value column doesn't end in _D, add it
        original_value_column = value_column
        value_column = self._ensure_deflator_suffix(value_column)

        # Rebase the deflator values
        rebased[value_column] = (
            100 * rebased[original_value_column] / rebased["base_year_value"]
        ).round(6)

        # Update the deflator data
        self.deflator_data = rebased.drop(columns=["base_year_value"])


class ExchangeDeflator(Deflator):
    """Manages and processes exchange rate deflators.

    Using exchange rate data (local currency to USD) provided by the source.
    It normalizes exchange rate deflators to a specified base year, allowing
    consistent comparisons over time.

    Args:
        source (Source): The source from which to fetch exchange rate data.
        base_year (int): The base year to rebase the deflator data.
    """

    def __init__(self, source: Exchange, base_year: int):
        super().__init__(
            source=source,
            deflator_type="exchange",
            price_kind=None,
            base_year=base_year,
        )


class PriceDeflator(Deflator):
    """Manages and processes price deflators data.

    It fetches price deflator data from the provided source and normalizes it to a specified
    base year, allowing for inflation-adjusted comparisons over time.

    Args:
        source (Source): The source from which to fetch price deflator data.
        kind (AvailableDeflators): The type of price deflator to apply (e.g., GDP deflator or CPI).
        base_year (int): The base year to rebase the deflator data.
    """

    def __init__(
        self,
        source: Source,
        kind: AvailableDeflators,
        base_year: int,
    ):
        super().__init__(
            source=source,
            deflator_type="price",
            price_kind=kind,
            base_year=base_year,
        )



from dataclasses import dataclass, field

import pandas as pd

from pydeflate.sources.common import AvailableDeflators
from pydeflate.sources.dac import read_dac
from pydeflate.sources.imf import read_weo
from pydeflate.sources.world_bank import read_wb, read_wb_lcu_ppp, read_wb_usd_ppp


@dataclass
class Source:
    name: str
    reader: callable
    update: bool = False
    data: pd.DataFrame = field(default_factory=pd.DataFrame)
    _idx = ["pydeflate_year", "pydeflate_entity_code", "pydeflate_iso3"]

    def __post_init__(self):
        self.data = self.reader(self.update)
        self.validate()

    def validate(self):
        if self.data.empty:
            raise ValueError(f"No data found for {self.name}")

        # check all columns start with pydeflate_
        if not all(col.startswith("pydeflate_") for col in self.data.columns):
            raise ValueError(f"Invalid data format for {self.name}")

    def lcu_usd_exchange(self) -> pd.DataFrame:
        return self.data.filter(self._idx + ["pydeflate_EXCHANGE"])

    def price_deflator(self, kind: AvailableDeflators = "NGDP_D") -> pd.DataFrame:

        if f"pydeflate_{kind}" not in self.data.columns:
            raise ValueError(f"No deflator data found for {kind} in {self.name} data.")

        return self.data.filter(self._idx + [f"pydeflate_{kind}"])


class IMF(Source):
    def __init__(self, update: bool = False):
        super().__init__(name="IMF", reader=read_weo, update=update)


class WorldBank(Source):
    def __init__(self, update: bool = False):
        super().__init__(name="World Bank", reader=read_wb, update=update)


class WorldBankPPP(Source):
    def __init__(self, update: bool = False, *, from_lcu: bool = True):
        super().__init__(
            name="World Bank PPP",
            reader=read_wb_lcu_ppp if from_lcu else read_wb_usd_ppp,
            update=update,
        )


class DAC(Source):
    def __init__(self, update: bool = False):
        super().__init__(name="DAC", reader=read_dac, update=update)



from functools import wraps

import pandas as pd

from pydeflate.core.api import BaseDeflate
from pydeflate.core.source import DAC, WorldBank, IMF


def _generate_docstring(source_name: str, price_kind: str) -> str:
    """Generate docstring for each decorated deflation function."""
    return (
        f"Deflate a DataFrame using the {source_name} deflator source ({price_kind}).\n\n"
        f"This function applies deflation adjustments to a DataFrame using the {source_name} {price_kind} deflator.\n\n"
        "Args:\n"
        "    data (pd.DataFrame): The input DataFrame containing data to deflate.\n"
        "    base_year (int): The base year for calculating deflation adjustments.\n"
        "    source_currency (str, optional): The source currency code. Defaults to 'USA'.\n"
        "    target_currency (str, optional): The target currency code. Defaults to 'USA'.\n"
        "    id_column (str, optional): Column with entity identifiers. Defaults to 'iso_code'.\n"
        "    year_column (str, optional): Column with year information. Defaults to 'year'.\n"
        "    use_source_codes (bool, optional): Use source-specific entity codes. Defaults to False.\n"
        "    value_column (str, optional): Column with values to deflate. Defaults to 'value'.\n"
        "    target_value_column (str, optional): Column to store deflated values. Defaults to 'value'.\n"
        "    to_current (bool, optional): Adjust values to current-year values if True. Defaults to False.\n"
        "    year_format (str | None, optional): Format of the year in `year_column`. Defaults to None.\n"
        "    update_deflators (bool, optional): Update the deflator data before deflation. Defaults to False.\n\n"
        "Returns:\n"
        "    pd.DataFrame: DataFrame with deflated values in the `target_value_column`.\n"
    )


def _deflator(deflator_source_cls, price_kind):
    """Decorator to create deflate wrappers with specific deflator source and price kind."""

    def decorator(func):
        @wraps(func)
        def wrapper(
            data: pd.DataFrame,
            *,
            base_year: int,
            source_currency: str = "USA",
            target_currency: str = "USA",
            id_column: str = "iso_code",
            year_column: str = "year",
            use_source_codes: bool = False,
            value_column: str = "value",
            target_value_column: str = "value",
            to_current: bool = False,
            year_format: str | None = None,
            update_deflators: bool = False,
        ):
            # Validate input parameters
            if not isinstance(data, pd.DataFrame):
                raise ValueError("The 'data' parameter must be a pandas DataFrame.")
            if not isinstance(base_year, int):
                raise ValueError("The 'base_year' parameter must be an integer.")
            if id_column not in data.columns:
                raise ValueError(
                    f"The id_column '{id_column}' is not in the DataFrame."
                )
            if year_column not in data.columns:
                raise ValueError(
                    f"The year_column '{year_column}' is not in the DataFrame."
                )
            if value_column not in data.columns:
                raise ValueError(
                    f"The value_column '{value_column}' is not in the DataFrame."
                )

            # Copy the data to avoid modifying the original
            to_deflate = data.copy()

            # Initialize the deflator source
            source = deflator_source_cls(update=update_deflators)

            # Create a deflator object
            deflator = BaseDeflate(
                base_year=base_year,
                deflator_source=source,
                exchange_source=source,
                source_currency=source_currency,
                target_currency=target_currency,
                price_kind=price_kind,
                use_source_codes=use_source_codes,
                to_current=to_current,
            )

            # Deflate the data
            return deflator.deflate(
                data=to_deflate,
                entity_column=id_column,
                year_column=year_column,
                value_column=value_column,
                target_value_column=target_value_column,
                year_format=year_format,
            )

        # Add the deflator source and price kind to the function
        wrapper.__doc__ = _generate_docstring(deflator_source_cls.__name__, price_kind)
        return wrapper

    return decorator


@_deflator(DAC, "NGDP_D")
def oecd_dac_deflate(
    data: pd.DataFrame,
    *,
    base_year: int,
    source_currency: str = "USA",
    target_currency: str = "USA",
    id_column: str = "iso_code",
    year_column: str = "year",
    use_source_codes: bool = False,
    value_column: str = "value",
    target_value_column: str = "value",
    to_current: bool = False,
    year_format: str | None = None,
    update_deflators: bool = False,
) -> pd.DataFrame: ...


@_deflator(WorldBank, "NGDP_D")
def wb_gdp_deflate(
    data: pd.DataFrame,
    *,
    base_year: int,
    source_currency: str = "USA",
    target_currency: str = "USA",
    id_column: str = "iso_code",
    year_column: str = "year",
    use_source_codes: bool = False,
    value_column: str = "value",
    target_value_column: str = "value",
    to_current: bool = False,
    year_format: str | None = None,
    update_deflators: bool = False,
): ...


@_deflator(WorldBank, "NGDP_D")
def wb_gdp_linked_deflate(
    data: pd.DataFrame,
    *,
    base_year: int,
    source_currency: str = "USA",
    target_currency: str = "USA",
    id_column: str = "iso_code",
    year_column: str = "year",
    use_source_codes: bool = False,
    value_column: str = "value",
    target_value_column: str = "value",
    to_current: bool = False,
    year_format: str | None = None,
    update_deflators: bool = False,
): ...


@_deflator(WorldBank, "CPI")
def wb_cpi_deflate(
    data: pd.DataFrame,
    *,
    base_year: int,
    source_currency: str = "USA",
    target_currency: str = "USA",
    id_column: str = "iso_code",
    year_column: str = "year",
    use_source_codes: bool = False,
    value_column: str = "value",
    target_value_column: str = "value",
    to_current: bool = False,
    year_format: str | None = None,
    update_deflators: bool = False,
): ...


@_deflator(IMF, "NGDP_D")
def imf_gdp_deflate(
    data: pd.DataFrame,
    *,
    base_year: int,
    source_currency: str = "USA",
    target_currency: str = "USA",
    id_column: str = "iso_code",
    year_column: str = "year",
    use_source_codes: bool = False,
    value_column: str = "value",
    target_value_column: str = "value",
    to_current: bool = False,
    year_format: str | None = None,
    update_deflators: bool = False,
): ...


@_deflator(IMF, "PCPI")
def imf_cpi_deflate(
    data: pd.DataFrame,
    *,
    base_year: int,
    source_currency: str = "USA",
    target_currency: str = "USA",
    id_column: str = "iso_code",
    year_column: str = "year",
    use_source_codes: bool = False,
    value_column: str = "value",
    target_value_column: str = "value",
    to_current: bool = False,
    year_format: str | None = None,
    update_deflators: bool = False,
): ...


@_deflator(IMF, "PCPIE")
def imf_cpi_e_deflate(
    data: pd.DataFrame,
    *,
    base_year: int,
    source_currency: str = "USA",
    target_currency: str = "USA",
    id_column: str = "iso_code",
    year_column: str = "year",
    use_source_codes: bool = False,
    value_column: str = "value",
    target_value_column: str = "value",
    to_current: bool = False,
    year_format: str | None = None,
    update_deflators: bool = False,
): ...



from concurrent.futures import ThreadPoolExecutor, as_completed
from pathlib import Path

import pandas as pd
import wbgapi as wb

from pydeflate.pydeflate_config import PYDEFLATE_PATHS, logger
from pydeflate.sources.common import (
    enforce_pyarrow_types,
    today,
    compute_exchange_deflator,
    read_data,
    prefix_pydeflate_to_columns,
)
from pydeflate.utils import emu

_INDICATORS: dict = {
    "NY.GDP.DEFL.ZS": "NGDP_D",  # GDP Deflator (Index)
    "NY.GDP.DEFL.ZS.AD": "NGDP_DL",  # GDP Deflator linked series
    "FP.CPI.TOTL": "CPI",  # Consumer Price Index (CPI)
    "PA.NUS.FCRF": "EXCHANGE",  # Official Exchange Rate
}

_INDICATORS_LCU_PPP: dict = {
    "NY.GDP.DEFL.ZS": "NGDP_D",  # GDP Deflator (Index)
    "NY.GDP.DEFL.ZS.AD": "NGDP_DL",  # GDP Deflator linked series
    "PA.NUS.PPP": "EXCHANGE",  # PPP conversion factor
}

_INDICATORS_USD_PPP: dict = {
    "NY.GDP.DEFL.ZS": "NGDP_D",  # GDP Deflator (Index)
    "NY.GDP.DEFL.ZS.AD": "NGDP_DL",  # GDP Deflator linked series
    "PA.NUS.PPPC.RF": "EXCHANGE",  # PPP conversion factor to market exchange rate
}


def get_wb_indicator(series: str, value_name: str | None = None) -> pd.DataFrame:
    """Fetch a World Bank indicator and transform it into a cleaned DataFrame.

    Args:
        series (str): The World Bank indicator series code.
        value_name (str | None): The column name to assign to the series values.
                                 If None, the series code will be used as the column name.

    Returns:
        pd.DataFrame: DataFrame with entity code, entity name, year, and the indicator values.
    """
    # Fetch the indicator data from World Bank API, clean and structure it
    return (
        wb.data.DataFrame(
            series=series,
            db=2,  # World Development Indicators database
            skipBlanks=True,
            columns="series",
            numericTimeKeys=True,
            labels=True,
        )
        .reset_index()
        .sort_values(by=["economy", "Time"])  # Sort for easier reading
        .drop(columns=["Time"])  # Remove unnecessary column
        .rename(
            columns={
                "economy": "entity_code",
                "Country": "entity",
                "time": "year",
                series: value_name or series,
            }
        )
        .reset_index(drop=True)  # Drop the old index after reset
    )


def _eur_series_fix(df: pd.DataFrame) -> pd.DataFrame:
    """
    Fix the exchange rate for the Euro area countries. This is done by assigning the
    exchange rate of the Euro to the countries in the Euro area. This is necessary
    because the series for Euro area countries are missing EUR exchange rates.

    Args:
        df: pd.DataFrame: The DataFrame containing the World Bank data.

    Returns:
        pd.DataFrame: The DataFrame with the fixed exchange rates for the Euro area countries.

    """
    # Handle cases where EUR is represented differently in the World Bank data
    df["entity_code"] = df["entity_code"].replace({"EMU": "EUR"})

    # Find the "Euro" data. This is done given that some countries are missing
    # exchange rates, but they are part of the Euro area.
    eur = (
        df.loc[lambda d: d["entity_code"] == "EUR"]
        .dropna(subset=["EXCHANGE"])
        .set_index("year")["EXCHANGE"]
        .to_dict()
    )

    # Euro area countries without exchange rates
    eur_mask = (df["entity_code"].isin(emu())) & (df["EXCHANGE"].isna())

    # Assign EURO exchange rate to euro are countries from year euro adopted
    df.loc[eur_mask, "EXCHANGE"] = df["year"].map(eur)

    return df


def _parallel_download_indicators(indicators: dict) -> list[pd.DataFrame]:
    """Download multiple World Bank indicators in parallel.

    Args:
        indicators (dict): A dictionary of World Bank indicators to download.

    Returns:
        list[pd.DataFrame]: A list of DataFrames containing the downloaded indicators

    """
    # List to store the resulting dataframes
    dfs = []

    # Use ThreadPoolExecutor to fetch indicators in parallel
    with ThreadPoolExecutor() as executor:
        # Submit all tasks to the executor (downloading indicators in parallel)
        future_to_series = {
            executor.submit(get_wb_indicator, series, value_name): series
            for series, value_name in indicators.items()
        }

        # Collect the results as they complete
        for future in as_completed(future_to_series):
            series = future_to_series[future]
            try:
                df_ = future.result().set_index(["year", "entity_code", "entity"])
                dfs.append(df_)
            except Exception as exc:
                # Log or handle any errors that occur during the download
                logger.warning(f"Error downloading series {series}: {exc}")

    return dfs


def _add_ppp_ppp_exchange(df: pd.DataFrame) -> pd.DataFrame:
    """
    Add the PPP exchange rate to the DataFrame.

    Args:
        df: pd.DataFrame: The DataFrame containing the World Bank data.

    Returns:
        pd.DataFrame: The DataFrame with the PPP exchange rates

    """
    ppp = df.loc[lambda d: d["entity_code"] == "USA"].copy()
    ppp[["entity_code", "entity", "pydeflate_iso3"]] = "PPP"

    df = pd.concat([df, ppp], ignore_index=True)

    return df


def _download_wb(
    indicators: dict, prefix: str = "wb", add_ppp_exchange: bool = False
) -> None:
    """Download multiple World Bank indicators in parallel and save as a parquet file.

    This function fetches all indicators defined in _INDICATORS in parallel, concatenates
    them into a single DataFrame, and saves the result as a parquet file using today's date as a suffix.
    """
    logger.info("Downloading the latest World Bank data...")

    indicators_data = _parallel_download_indicators(indicators=indicators)

    # Concatenate all DataFrames horizontally (by columns)
    df = pd.concat(indicators_data, axis=1).reset_index()

    # cleaning
    df = (
        df.pipe(_eur_series_fix)
        .pipe(compute_exchange_deflator, base_year_measure="NGDP_D")
        .assign(pydeflate_iso3=lambda d: d.entity_code)
        .sort_values(by=["year", "entity_code"])
    )

    if add_ppp_exchange:
        df = df.pipe(_add_ppp_ppp_exchange)

    df = (
        df.pipe(prefix_pydeflate_to_columns)
        .pipe(enforce_pyarrow_types)
        .reset_index(drop=True)
    )

    # Get today's date to use as a file suffix
    suffix = today()

    # Save the DataFrame as a parquet file
    output_path = PYDEFLATE_PATHS.data / f"{prefix}_{suffix}.parquet"
    df.to_parquet(output_path)

    logger.info(f"Saved World Bank data to {prefix}_{suffix}.parquet")


def download_wb() -> None:
    """Download the latest World Bank data."""
    _download_wb(indicators=_INDICATORS, prefix="wb")


def download_wb_lcu_ppp() -> None:
    """Download the latest World Bank data (PPP)."""
    _download_wb(
        indicators=_INDICATORS_LCU_PPP, prefix="wb_lcu_ppp", add_ppp_exchange=True
    )


def download_wb_usd_ppp() -> None:
    """Download the latest World Bank data (PPP)."""
    _download_wb(
        indicators=_INDICATORS_USD_PPP, prefix="wb_usd_ppp", add_ppp_exchange=True
    )


def _find_wb_files_in_path(path: Path) -> list:
    """Find all WB parquet files in the specified directory.

    Args:
        path (Path): The directory path to search for WB parquet files.

    Returns:
        list: List of WB parquet files found in the directory.
    """
    return list(path.glob(f"wb_*.parquet"))


def _find_wb_lcu_ppp_files_in_path(path: Path) -> list:
    """Find all WB PPP parquet files in the specified directory.

    Args:
        path (Path): The directory path to search for WB parquet files.

    Returns:
        list: List of WB parquet files found in the directory.
    """
    return list(path.glob(f"wb_lcu_ppp_*.parquet"))


def _find_wb_usd_ppp_files_in_path(path: Path) -> list:
    """Find all WB PPP parquet files in the specified directory.

    Args:
        path (Path): The directory path to search for WB parquet files.

    Returns:
        list: List of WB parquet files found in the directory.
    """
    return list(path.glob(f"wb_usd_ppp_*.parquet"))


def read_wb(update: bool = False) -> pd.DataFrame:
    """Read the latest World Bank data from parquet files or download fresh data."""
    return read_data(
        file_finder_func=_find_wb_files_in_path,
        download_func=download_wb,
        data_name="World Bank",
        update=update,
    )


def read_wb_lcu_ppp(update: bool = False) -> pd.DataFrame:
    """Read the latest World Bank data from parquet files or download fresh data."""
    return read_data(
        file_finder_func=_find_wb_lcu_ppp_files_in_path,
        download_func=download_wb_lcu_ppp,
        data_name="World Bank",
        update=update,
    )


def read_wb_usd_ppp(update: bool = False) -> pd.DataFrame:
    """Read the latest World Bank data from parquet files or download fresh data."""
    return read_data(
        file_finder_func=_find_wb_usd_ppp_files_in_path,
        download_func=download_wb_usd_ppp,
        data_name="World Bank",
        update=update,
    )


if __name__ == "__main__":
    df_wb = read_wb(False)
    df_usd = read_wb_usd_ppp(False)
    df_lcu = read_wb_lcu_ppp(False)



import pandas as pd

from pydeflate.core.deflator import ExchangeDeflator, PriceDeflator
from pydeflate.core.exchange import Exchange
from pydeflate.core.source import Source
from pydeflate.sources.common import AvailableDeflators
from pydeflate.utils import (
    create_pydeflate_year,
    merge_user_and_pydeflate_data,
    get_unmatched_pydeflate_data,
    get_matched_pydeflate_data,
    flag_missing_pydeflate_data,
)


def resolve_common_currencies(currency: str, source: str) -> str:
    mapping = {
        "USD": "USA",
        "EUR": "EMU",
        "GBP": "GBR",
        "JPY": "JPN",
        "CAD": "CAN",
    }

    if source == "DAC":
        mapping["EUR"] = "EUI"

    return mapping.get(currency, currency)


def _base_operation(
    base_obj,
    data: pd.DataFrame,
    entity_column: str,
    year_column: str,
    value_column: str,
    target_value_column: str | None = None,
    year_format: str | None = None,
    exchange: bool = False,
    reversed_: bool = False,
):
    """Perform deflation or exchange rate adjustment on input data using pydeflate data.

    Args:
        base_obj (BaseExchange | BaseDeflate): The base object containing pydeflate data.
        data (pd.DataFrame): Data to be adjusted.
        entity_column (str): Column with entity or country identifiers.
        year_column (str): Column with year information.
        value_column (str): Column with values to be adjusted.
        target_value_column (str | None, optional): Column to store adjusted values. Defaults to `value_column`.
        year_format (str, optional): Format of the year. Defaults to "%Y".
        exchange (bool, optional): Whether to perform an exchange rate adjustment (True)
        or deflation (False).
        reversed_ (bool, optional): If True, perform the operation in reverse.

    Returns:
        pd.DataFrame: DataFrame with adjusted values and original columns preserved.
    """
    # Make a copy of the input data to avoid modifying the original data
    data = data.copy(deep=True)

    target_value_column = target_value_column or value_column

    # Keep track of original columns to return data in the same order.
    cols = (
        [*data.columns, target_value_column]
        if target_value_column not in data.columns
        else data.columns
    )

    # Merge pydeflate data to the input data
    base_obj._merge_pydeflate_data(
        data=data,
        entity_column=entity_column,
        year_column=year_column,
        year_format=year_format,
    )

    # Flag missing data
    flag_missing_pydeflate_data(
        base_obj._unmatched_data, entity_column=entity_column, year_column=year_column
    )
    x = base_obj._merged_data[value_column]
    y = base_obj._merged_data[
        "pydeflate_EXCHANGE" if exchange else "pydeflate_deflator"
    ]

    # Apply the correct operation based on `exchange` and `reversed`
    if (exchange and not reversed_) or (not exchange and reversed_):
        base_obj._merged_data[target_value_column] = (x * y).round(6)
    else:
        base_obj._merged_data[target_value_column] = (x / y).round(6)

    return base_obj._merged_data[cols]


class BaseExchange:
    """Performs currency exchange rate conversions within data, using pydeflate.

    Provides methods to merge pydeflate exchange rate data with user data and apply
    exchange rate adjustments, allowing for straightforward currency conversions
    within a dataset.
    """

    def __init__(
        self,
        exchange_source: Source,
        source_currency: str,
        target_currency: str,
        use_source_codes: bool = False,
    ):
        """Initialize a BaseExchange instance with exchange rate data and identifiers.

        Args:
            exchange_source (Source): Data source for exchange rates.
            source_currency (str): Original currency for conversion.
            target_currency (str): Target currency for conversion.
            use_source_codes (bool, optional): Use source-specific entity codes.
             Defaults to False.
        """

        # Try to accept common currencies by their country codes
        source_currency = resolve_common_currencies(
            source_currency, exchange_source.name
        )

        target_currency = resolve_common_currencies(
            target_currency, exchange_source.name
        )

        self.exchange_rates = Exchange(
            source=exchange_source,
            source_currency=source_currency,
            target_currency=target_currency,
        )

        self._idx = [
            "pydeflate_year",
            "pydeflate_entity_code" if use_source_codes else "pydeflate_iso3",
        ]

        self._load_pydeflate_data()

    def _load_pydeflate_data(self) -> None:
        """Load and prepare pydeflate exchange rate data."""
        self.pydeflate_data = (
            self.exchange_rates.exchange_data.set_index(self._idx)
            .dropna(how="any")
            .reset_index()
        )

    def _merge_pydeflate_data(
        self,
        data: pd.DataFrame,
        entity_column: str,
        year_column: str,
        year_format: str | None = None,
    ) -> None:
        """Merge pydeflate exchange rate data into input data by year and entity.

        Args:
            data (pd.DataFrame): Input DataFrame to merge with pydeflate data.
            entity_column (str): Column name for entity or country codes.
            year_column (str): Column name for year information.
            year_format (str, optional): Year format. Defaults to '%Y'.
        """

        # Convert the year to an integer
        data = create_pydeflate_year(
            data=data, year_column=year_column, year_format=year_format
        )

        # Merge data to the input data based on year and entity
        merged_data = merge_user_and_pydeflate_data(
            data=data,
            pydeflate_data=self.pydeflate_data,
            entity_column=entity_column,
            ix=self._idx,
            source_codes=self._idx[-1] == "pydeflate_entity_code",
            dac=self.exchange_rates.source.name == "DAC",
        )

        # store unmatched data
        self._unmatched_data = get_unmatched_pydeflate_data(merged_data=merged_data)

        # store matched data
        self._merged_data = get_matched_pydeflate_data(merged_data=merged_data)

    def exchange(
        self,
        data: pd.DataFrame,
        entity_column: str,
        year_column: str,
        value_column: str,
        target_value_column: str | None = None,
        year_format: str | None = None,
        reversed_: bool = False,
    ):
        """Apply exchange rate conversion to input data.

        Args:
            data (pd.DataFrame): Data to apply exchange rate adjustment.
            entity_column (str): Column with entity identifiers.
            year_column (str): Column with year information.
            value_column (str): Column with values to adjust.
            target_value_column (str | None, optional): Column for adjusted values. Defaults to `value_column`.
            year_format (str, optional): Format of the year. Defaults to "%Y".
            reversed_ (bool, optional): If True, perform the operation in reverse. defaults to False.

        Returns:
            pd.DataFrame: DataFrame with exchange rate-adjusted values.
        """
        return _base_operation(
            base_obj=self,
            exchange=True,
            data=data,
            entity_column=entity_column,
            year_column=year_column,
            value_column=value_column,
            target_value_column=target_value_column,
            year_format=year_format,
            reversed_=reversed_,
        )


class BaseDeflate:
    """Deflating monetary values over time with exchange rates and price indices.

    Combines price deflators, exchange rate deflators, and pydeflate data to allow
    for adjustments of monetary values across different time periods, enabling accurate
    economic comparisons over time.
    """

    def __init__(
        self,
        deflator_source: Source,
        exchange_source: Source,
        base_year: int,
        price_kind: AvailableDeflators,
        source_currency: str,
        target_currency: str,
        use_source_codes: bool = False,
        to_current: bool = False,
    ):
        """Initialize a BaseDeflate instance with deflator and exchange rate sources.

        Args:
            deflator_source (Source): Data source for price deflators.
            exchange_source (Source): Data source for exchange rates.
            base_year (int): Reference year for base value adjustments.
            price_kind (AvailableDeflators): Type of deflator (e.g., CPI).
            source_currency (str): Currency of the input data.
            target_currency (str): Currency for conversion.
            use_source_codes (bool, optional): Use source-specific entity codes. Defaults to False.
            to_current (bool, optional): If True, adjust to current year values. Defaults to False.
        """

        # Try to accept common currencies by their country codes
        source_currency = resolve_common_currencies(
            source_currency, deflator_source.name
        )

        target_currency = resolve_common_currencies(
            target_currency, deflator_source.name
        )
        self.exchange_rates = Exchange(
            source=exchange_source,
            source_currency=source_currency,
            target_currency=target_currency,
        )

        self.exchange_deflator = ExchangeDeflator(
            source=self.exchange_rates, base_year=base_year
        )

        self.price_deflator = PriceDeflator(
            source=deflator_source, kind=price_kind, base_year=base_year
        )

        self._idx = [
            "pydeflate_year",
            "pydeflate_entity_code" if use_source_codes else "pydeflate_iso3",
        ]

        self.use_source_codes = use_source_codes
        self.to_current = to_current

        self.__post_init__()

    def __post_init__(self):
        """Post-initialization process to merge deflator, exchange, and pydeflate data."""

        # Merge deflator and exchange rate data
        data = self._merge_components(
            df=self.price_deflator.deflator_data,
            other=self.exchange_deflator.deflator_data,
        ).pipe(self._merge_components, other=self.exchange_rates.exchange_data)

        # drop where necessary data is missing
        data = data.set_index(self._idx).dropna(how="any").reset_index()

        # Calculate price-exchange deflator
        data["pydeflate_deflator"] = self._calculate_deflator_value(
            data[f"pydeflate_{self.price_deflator.price_kind}"],
            data["pydeflate_EXCHANGE_D"],
            data[f"pydeflate_EXCHANGE"],
        )

        self.pydeflate_data = data

    def _calculate_deflator_value(
        self, price_def: pd.Series, exchange_def: pd.Series, exchange_rate: pd.Series
    ):
        """Compute the combined deflator value using price deflator, exchange deflator, and rates.

        Args:
            price_def (pd.Series): Series of price deflator values.
            exchange_def (pd.Series): Series of exchange deflator values.
            exchange_rate (pd.Series): Series of exchange rates.

        Returns:
            pd.Series: Series with combined deflator values.
        """
        return (
            (exchange_def * exchange_rate) / price_def
            if self.to_current
            else price_def / (exchange_def * exchange_rate)
        )

    def _merge_components(self, df: pd.DataFrame, other: pd.DataFrame):
        """Combine data components, merging deflator and exchange rate information.

        Args:
            df (pd.DataFrame): Main DataFrame for merging.
            other (pd.DataFrame): Additional data to merge into `df`.

        Returns:
            pd.DataFrame: Merged DataFrame without duplicate columns.
        """
        merged = df.merge(other, how="outer", on=self._idx, suffixes=("", "_ex"))
        return merged.drop(columns=merged.filter(regex=f"_ex$").columns, axis=1)

    def _merge_pydeflate_data(
        self,
        data: pd.DataFrame,
        entity_column: str,
        year_column: str,
        year_format: str | None = None,
    ) -> None:
        """Merge pydeflate deflator data into the input data by year and entity.

        Args:
            data (pd.DataFrame): Input DataFrame to merge with pydeflate data.
            entity_column (str): Column for entity or country identifiers.
            year_column (str): Column for year information.
            year_format (str, optional): Format of the year. Defaults to '%Y'.
        """

        # Convert the year to an integer
        data = create_pydeflate_year(
            data=data, year_column=year_column, year_format=year_format
        )

        # Merge data to the input data based on year and entity
        merged_data = merge_user_and_pydeflate_data(
            data=data,
            pydeflate_data=self.pydeflate_data,
            entity_column=entity_column,
            ix=self._idx,
            source_codes=self.use_source_codes,
            dac=self.exchange_deflator.source.name == "DAC"
            or self.price_deflator.source.name == "DAC",
        )

        # store unmatched data
        self._unmatched_data = get_unmatched_pydeflate_data(merged_data=merged_data)

        # store matched data
        self._merged_data = get_matched_pydeflate_data(merged_data=merged_data)

    def deflate(
        self,
        data: pd.DataFrame,
        entity_column: str,
        year_column: str,
        value_column: str,
        target_value_column: str | None = None,
        year_format: str | None = None,
    ):
        """Apply deflation adjustment to input data using pydeflate deflator rates.

        Args:
            data (pd.DataFrame): Data for deflation adjustment.
            entity_column (str): Column with entity identifiers.
            year_column (str): Column with year information.
            value_column (str): Column with values to deflate.
            target_value_column (str | None, optional): Column to store deflated values. Defaults to `value_column`.
            year_format (str, optional): Format of the year. Defaults to "%Y".

        Returns:
            pd.DataFrame: DataFrame with deflated values.
        """
        return _base_operation(
            base_obj=self,
            data=data,
            entity_column=entity_column,
            year_column=year_column,
            value_column=value_column,
            target_value_column=target_value_column,
            year_format=year_format,
        )



import logging
from pathlib import Path


class PYDEFLATE_PATHS:
    """Class to store the paths to the data and output folders."""

    package = Path(__file__).resolve().parent.parent
    data = package / "pydeflate" / ".pydeflate_data"
    settings = package / "pydeflate" / "settings"
    test_data = package / "tests" / "test_files"


def setup_logger(name) -> logging.Logger:
    """Set up the logger.

    Args:
        name (str): The name of the logger.

    Returns:
        logging.Logger: The logger.

    """
    logger_ = logging.getLogger(name)
    logger_.setLevel(logging.INFO)

    # Only add handlers if the logger has none to avoid duplication
    if not logger_.hasHandlers():
        console_handler = logging.StreamHandler()
        console_handler.setLevel(logging.INFO)

        formatter = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s:\n %(message)s"
        )
        console_handler.setFormatter(formatter)

        logger_.addHandler(console_handler)
        logger_.propagate = False

    return logger_


logger = setup_logger("pydeflate")



from dataclasses import dataclass, field

import pandas as pd

from pydeflate.core.source import Source
from pydeflate.sources.common import compute_exchange_deflator


@dataclass
class Exchange:
    """A class to manage and process exchange rate data for currency conversions.

    This class holds exchange rate data and provides methods to apply exchange rates
    for conversions between source and target currencies.

    Attributes:
        source (Source): An instance of the Source class to fetch exchange rate data.
        source_currency (str): The source currency code (default is 'LCU' - Local Currency Unit).
        target_currency (str): The target currency code (default is 'USA' - US Dollar).
        exchange_data (pd.DataFrame): DataFrame holding the exchange rate data.
    """

    source: Source
    source_currency: str = "LCU"
    target_currency: str = "USA"
    exchange_data: pd.DataFrame = field(default_factory=pd.DataFrame)

    def __post_init__(self):
        """Initialize the Exchange object and process the exchange rate data."""
        # Load and filter the relevant columns from the exchange rate data
        self.exchange_data = self.source.lcu_usd_exchange()

        if self.source_currency == self.target_currency:
            self.exchange_data = self.exchange_rate("LCU", self.target_currency)
            self.exchange_data["pydeflate_EXCHANGE"] = 1
        else:
            self.exchange_data = self.exchange_rate(
                self.source_currency, self.target_currency
            )

    def _get_exchange_rate(self, currency):
        """Helper function to fetch exchange rates for a given currency."""
        exchange_rate = self.exchange_data.loc[
            self.exchange_data["pydeflate_iso3"] == currency
        ]
        return exchange_rate

    def _convert_exchange(self, to_: str) -> pd.DataFrame:
        """Converts exchange rates based on the target currency.

        This method retrieves exchange rates for a given target currency, merges the
        target exchange rates with the base exchange rates, and computes the final
        exchange rate by dividing the base exchange rate by the target exchange rate.

        Args:
            to_ (str): The target currency code.

        Returns:
            pd.DataFrame: A DataFrame with the adjusted exchange rates for the target currency.

        Raises:
            ValueError: If no exchange rate data is available for the target currency.
        """

        if to_ == "LCU":
            return self.exchange_data.copy().assign(pydeflate_EXCHANGE=1)

        usd_rate = self.exchange_data.copy()
        target_exchange = self._get_exchange_rate(to_)

        if target_exchange.empty:
            raise ValueError(f"No currency exchange data for {to_=}")

        merged = pd.merge(
            usd_rate,
            target_exchange,
            how="left",
            on=["pydeflate_year"],
            suffixes=("", "_to"),
        ).assign(
            pydeflate_EXCHANGE=lambda d: d.pydeflate_EXCHANGE / d.pydeflate_EXCHANGE_to,
        )

        return merged.drop(columns=merged.filter(regex="_to$").columns, axis=1)

    def exchange_rate(self, from_currency: str, to_currency: str):
        """Calculates the exchange rates between the source and target currencies.

        Args:
            from_currency (str): The source currency code.
            to_currency (str): The target currency code.

        Returns:
            None
        """
        # Get exchange data based on the source currency.
        source = self._convert_exchange(to_=from_currency)

        # Get exchange data based on the target currency.
        target = self._convert_exchange(to_=to_currency)

        # Merge the source and target exchange data and compute the final exchange rate.
        merged = pd.merge(
            source,
            target,
            how="left",
            on=["pydeflate_year", "pydeflate_entity_code", "pydeflate_iso3"],
            suffixes=("", "_to"),
        ).assign(
            pydeflate_EXCHANGE=lambda d: d.pydeflate_EXCHANGE / d.pydeflate_EXCHANGE_to
        )

        # Compute the exchange rate deflator
        merged = compute_exchange_deflator(
            merged,
            exchange="pydeflate_EXCHANGE_to",
            year="pydeflate_year",
            grouper=["pydeflate_entity_code", "pydeflate_iso3"],
        )

        # Drop unnecessary columns
        merged = merged.drop(columns=merged.filter(regex="_to$").columns, axis=1)

        return merged

    def exchange(
        self,
        data: pd.DataFrame,
        value_column: str,
        entity_column: str,
        year_column: str,
        year_format: str = "%Y",
        use_source_codes: bool = False,
    ) -> pd.DataFrame:
        """Apply the exchange rate to the given data to convert it to the target currency.

        Args:
            data (pd.DataFrame): The input DataFrame containing values to be exchanged.
            value_column (str): The column name containing the values to be converted.
            entity_column (str): The column name containing the entity or country codes.
            year_column (str): The column name containing the year information.
            year_format (str, optional): The format of the year (default is '%Y').
            use_source_codes (bool, optional): Whether to use source entity codes
            instead of ISO3 (default is False).

        Returns:
            pd.DataFrame: DataFrame with the values converted to the target currency.
        """

        # Convert the year to an integer
        data["pydeflate_year"] = pd.to_datetime(
            data[year_column], format=year_format
        ).dt.year

        # exchange columns
        exchange_cols = [
            "pydeflate_year",
            "pydeflate_entity_code",
            "pydeflate_iso3",
            "pydeflate_EXCHANGE",
        ]

        # Merge exchange rate data to the input data based on year and entity
        merged_data = data.merge(
            self.exchange_data.filter(exchange_cols),
            how="left",
            left_on=["pydeflate_year", entity_column],
            right_on=[
                "pydeflate_year",
                "pydeflate_entity_code" if use_source_codes else "pydeflate_iso3",
            ],
        )

        # Apply the exchange rate to convert the value column
        merged_data[value_column] = (
            merged_data[value_column] / merged_data["pydeflate_EXCHANGE"]
        )

        # Drop all columns that start with pydeflate_ merging and return the result
        return merged_data.filter(regex="^(?!pydeflate_)")

    def deflator(self) -> pd.DataFrame:
        """Get the exchange rate deflator data.

        Returns:
            pd.DataFrame: DataFrame with the exchange rate deflator data.
        """

        return self.exchange_data.filter(
            [
                "pydeflate_year",
                "pydeflate_entity_code",
                "pydeflate_iso3",
                "pydeflate_EXCHANGE_D",
            ]
        )

    def merge_deflator(
        self,
        data: pd.DataFrame,
        entity_column: str,
        year_column: str,
        year_format: str = "%Y",
        use_source_codes: bool = False,
    ) -> pd.DataFrame:
        """Merge the deflator data to the input data based on year and entity.

        Args:
            data (pd.DataFrame): The input DataFrame
            entity_column (str): The column name containing the entity or country codes.
            year_column (str): The column name containing the year information.
            year_format (str, optional): The format of the year (default is '%Y').
            use_source_codes (bool, optional): Whether to use source entity codes
            instead of ISO3 (default is False).

        Returns:
            pd.DataFrame: DataFrame with the deflator data merged to the input data.

        """

        # Convert the year to an integer
        data["pydeflate_year"] = pd.to_datetime(
            data[year_column], format=year_format
        ).dt.year

        # Merge exchange rate data to the input data based on year and entity
        merged_data = data.merge(
            self.deflator(),
            how="left",
            left_on=["pydeflate_year", entity_column],
            right_on=[
                "pydeflate_year",
                "pydeflate_entity_code" if use_source_codes else "pydeflate_iso3",
            ],
        )

        return merged_data.filter(regex="^(?!pydeflate_)")
